import { http as httpFollow } from 'follow-redirects';
import { https as httpsFollow } from 'follow-redirects';
import * as http from 'http';
import * as https from 'https';
import { Observable } from 'rxjs';
import * as url from 'url';
import * as zlib from 'zlib';
import { oxidVersion } from '../metadata';
import { isString } from '../utils/base';
import { isArrayBuffer, isStream } from '../utils/base';
import { createError, enhanceError } from '../utils/createError';
import { getObserverHandler } from '../utils/getObserverHandler';
import { parseJsonResponse } from '../utils/parseJsonResponse';
import { buildURL } from '../utils/urls';
var isHttps = /https:?/;
var httpAdapter = function (config) {
    return new Observable(function (observer) {
        var _a = getObserverHandler(observer), emitError = _a.emitError, emitComplete = _a.emitComplete;
        var transportRequest;
        var tearDown = function () {
            if (!transportRequest || transportRequest.aborted) {
                return;
            }
            transportRequest.abort();
        };
        var data = config.data;
        var proxy = config.proxy, _b = config.headers, headers = _b === void 0 ? {} : _b, maxContentLength = config.maxContentLength, configUrl = config.url, method = config.method;
        if (!configUrl || !method) {
            emitError(createError("Invalid request configuration"));
            return tearDown;
        }
        if (!headers['User-Agent'] && !headers['user-agent']) {
            headers['User-Agent'] = "oxid/" + oxidVersion;
        }
        if (data && !isStream(data)) {
            if (Buffer.isBuffer(data)) {
            }
            else if (isArrayBuffer(data)) {
                data = Buffer.from(new Uint8Array(data));
            }
            else if (isString(data)) {
                data = Buffer.from(data, 'utf-8');
            }
            else {
                emitError(createError('Data after transformation must be a string, an ArrayBuffer, a Buffer, or a Stream', config));
                return tearDown;
            }
            headers['Content-Length'] = data.length;
        }
        var auth = undefined;
        if (config.auth) {
            var username = config.auth.username || '';
            var password = config.auth.password || '';
            auth = username + ':' + password;
        }
        var parsed = url.parse(configUrl);
        var protocol = parsed.protocol || 'http:';
        if (!auth && parsed.auth) {
            var urlAuth = parsed.auth.split(':');
            var urlUsername = urlAuth[0] || '';
            var urlPassword = urlAuth[1] || '';
            auth = urlUsername + ':' + urlPassword;
        }
        if (auth) {
            delete headers.Authorization;
        }
        var isHttpsRequest = isHttps.test(protocol);
        var agent = isHttpsRequest ? config.httpsAgent : config.httpAgent;
        var options = {
            path: buildURL(parsed.path, config.params, config.paramsSerializer).replace(/^\?/, ''),
            method: method.toUpperCase(),
            headers: headers,
            agent: agent,
            auth: auth
        };
        if (config.socketPath) {
            options.socketPath = config.socketPath;
        }
        else {
            options.hostname = parsed.hostname;
            options.port = parsed.port;
        }
        if (proxy) {
            options.hostname = proxy.host;
            options.host = proxy.host;
            options.headers.host = parsed.hostname + (parsed.port ? ':' + parsed.port : '');
            options.port = proxy.port;
            options.path = protocol + '//' + parsed.hostname + (parsed.port ? ':' + parsed.port : '') + options.path;
            if (proxy.auth) {
                var base64 = Buffer.from(proxy.auth.username + ":" + proxy.auth.password, 'utf8').toString('base64');
                options.headers['Proxy-Authorization'] = 'Basic ' + base64;
            }
        }
        var transport;
        var isHttpsProxy = isHttpsRequest && (proxy ? isHttps.test(proxy.protocol || 'http') : true);
        if (config.transport) {
            transport = config.transport;
        }
        else if (config.maxRedirects === 0) {
            transport = isHttpsProxy ? https : http;
        }
        else {
            if (config.maxRedirects) {
                options.maxRedirects = config.maxRedirects;
            }
            transport = isHttpsProxy ? httpsFollow : httpFollow;
        }
        if (maxContentLength && maxContentLength > -1) {
            options.maxBodyLength = maxContentLength;
        }
        transportRequest = transport.request(options, function (res) {
            if (transportRequest.aborted) {
                return;
            }
            var stream = res;
            switch (res.headers['content-encoding']) {
                case 'gzip':
                case 'compress':
                case 'deflate':
                    stream = stream.pipe(zlib.createUnzip());
                    delete res.headers['content-encoding'];
                    break;
            }
            var lastRequest = res.req || transportRequest;
            var response = {
                status: Number.isSafeInteger(res.statusCode) ? res.statusCode : Number.NaN,
                statusText: res.statusMessage || '',
                headers: res.headers,
                config: config,
                request: lastRequest
            };
            if (config.responseType === "stream") {
                response.data = stream;
                emitComplete(response);
            }
            else {
                var responseBuffer_1 = [];
                stream.on('data', function (chunk) {
                    responseBuffer_1.push(chunk);
                    if (!!maxContentLength && maxContentLength > -1 && Buffer.concat(responseBuffer_1).length > maxContentLength) {
                        emitError(createError("maxContentLength size of " + maxContentLength + " exceeded", config, null, lastRequest));
                    }
                });
                stream.on('error', function (err) {
                    if (transportRequest.aborted) {
                        return;
                    }
                    emitError(enhanceError(err, config, null, lastRequest));
                });
                stream.on('end', function () {
                    var responseData = Buffer.concat(responseBuffer_1.filter(function (x) { return !!x; }));
                    if (config.responseType !== "arraybuffer") {
                        responseData = responseData.toString(config.responseEncoding);
                    }
                    var parsedBody = config.responseType === 'json' && isString(responseData)
                        ? parseJsonResponse(response.status, responseData)
                        : { ok: true, body: responseData };
                    response.data = parsedBody.body;
                    if (!parsedBody.ok) {
                        emitError(createError('Response parse failed', config, response.statusText, transportRequest, response));
                    }
                    else {
                        if (response.status === 0) {
                            response.status = !!responseData ? 200 : 0;
                        }
                        emitComplete(response);
                    }
                });
            }
        });
        transportRequest.on('error', function (err) {
            if (transportRequest.aborted) {
                return;
            }
            emitError(enhanceError(err, config, null, transportRequest));
        });
        if (isStream(data)) {
            data.on('error', function (err) { return emitError(enhanceError(err, config, null, transportRequest)); }).pipe(transportRequest);
        }
        else {
            transportRequest.end(data);
        }
        observer.next({ type: "Sent" });
        return tearDown;
    });
};
export { httpAdapter as adapter };
//# sourceMappingURL=http.js.map