import { Observable } from 'rxjs';
import { isFormData, isStandardBrowserEnv, isString, XSRF_HEADER_NAME } from '../utils/base';
import { cookies } from '../utils/cookies';
import { createError } from '../utils/createError';
import { getObserverHandler } from '../utils/getObserverHandler';
import { parseHeaders } from '../utils/parseHeaders';
import { parseJsonResponse } from '../utils/parseJsonResponse';
import { buildURL, isURLSameOrigin } from '../utils/urls';
import { xhrBackend } from './xhrBackend';
function getResponseUrl(xhr) {
    if ('responseURL' in xhr && xhr.responseURL) {
        return xhr.responseURL;
    }
    if (/^X-Request-URL:/m.test(xhr.getAllResponseHeaders())) {
        return xhr.getResponseHeader('X-Request-URL');
    }
    return null;
}
var xhrAdapter = function (config) {
    return new Observable(function (observer) {
        var reportProgress = config.reportProgress, withCredentials = config.withCredentials;
        var _a = getObserverHandler(observer), emitError = _a.emitError, emitComplete = _a.emitComplete;
        var xhr = xhrBackend();
        var tearDown = function () {
            xhr.removeEventListener('error', onError);
            xhr.removeEventListener('load', onLoad);
            xhr.removeEventListener('abort', onAbort);
            if (reportProgress) {
                xhr.removeEventListener('progress', onDownProgress);
                if (requestData !== null && xhr.upload) {
                    xhr.upload.removeEventListener('progress', onUpProgress);
                }
            }
            xhr.abort();
        };
        if (!config.url || !config.method) {
            emitError(createError("Invalid request configuration"));
            return tearDown;
        }
        var requestData = config.data;
        var requestHeaders = config.headers;
        if (isFormData(requestData)) {
            delete requestHeaders['Content-Type'];
        }
        if (config.auth) {
            var username = config.auth.username || '';
            var password = config.auth.password || '';
            requestHeaders.Authorization = 'Basic ' + btoa(username + ':' + password);
        }
        xhr.open(config.method.toUpperCase(), buildURL(config.url, config.params, config.paramsSerializer));
        if (!!withCredentials) {
            xhr.withCredentials = true;
        }
        if (isStandardBrowserEnv()) {
            var xsrfValue = (config.withCredentials || isURLSameOrigin(config.url)) && config.xsrfCookieName
                ? cookies.read(config.xsrfCookieName)
                : undefined;
            if (xsrfValue) {
                requestHeaders[config.xsrfHeaderName || XSRF_HEADER_NAME] = xsrfValue;
            }
        }
        if ('setRequestHeader' in xhr && !!requestHeaders) {
            Object.keys(requestHeaders)
                .map(function (key) { return [key, requestHeaders[key]]; })
                .filter(function (_a) {
                var key = _a[0];
                return key.toLowerCase() !== 'content-type' || !!requestData;
            })
                .forEach(function (_a) {
                var key = _a[0], value = _a[1];
                return xhr.setRequestHeader(key, value);
            });
            if (!requestHeaders['Accept']) {
                xhr.setRequestHeader('Accept', 'application/json, text/plain, */*');
            }
        }
        if (config.responseType) {
            try {
                var responseType = config.responseType.toLowerCase();
                xhr.responseType = (responseType !== "json" ? responseType : "text");
            }
            catch (e) {
                if (config.responseType !== 'json') {
                    throw e;
                }
            }
        }
        var headerResponse = null;
        var partialFromXhr = function () {
            if (headerResponse !== null) {
                return headerResponse;
            }
            var status = xhr.status;
            var statusText = xhr.statusText || 'OK';
            var headers = 'getAllResponseHeaders' in xhr ? parseHeaders(xhr.getAllResponseHeaders()) : null;
            var url = getResponseUrl(xhr) || config.url;
            headerResponse = { headers: headers, status: status, statusText: statusText, url: url, config: config, type: "ResponseHeader" };
            return headerResponse;
        };
        var onLoad = function () {
            var _a = partialFromXhr(), headers = _a.headers, status = _a.status, statusText = _a.statusText, url = _a.url;
            var body = null;
            if (status !== 204) {
                body = typeof xhr.response === 'undefined' ? xhr.responseText : xhr.response;
            }
            if (status === 0) {
                status = !!body ? 200 : 0;
            }
            var parsedBody = config.responseType === 'json' && isString(body) ? parseJsonResponse(status, body) : { ok: true, body: body };
            var responseData = !config.responseType || config.responseType === "text" ? xhr.responseText : xhr.response;
            var response = {
                type: "Response",
                data: parsedBody.body || responseData || null,
                status: status || xhr.status,
                statusText: statusText,
                headers: headers,
                config: config,
                request: xhr,
                responseUrl: url
            };
            if (!parsedBody.ok) {
                emitError(createError('Response parse failed', config, response.statusText, xhr, response));
            }
            else {
                emitComplete(response);
            }
        };
        var onError = function () {
            if (xhr.status === 0 && !!xhr.responseURL && xhr.responseURL.indexOf('file:') === 0) {
                return;
            }
            emitError(createError(xhr.statusText || 'Network Error', config, null, xhr));
        };
        var onAbort = function () { return emitError(createError('Request aborted', config, 'ECONNABORTED', xhr)); };
        var sentHeaders = false;
        var onDownProgress = function (event) {
            if (!sentHeaders) {
                observer.next(partialFromXhr());
                sentHeaders = true;
            }
            var progressEvent = {
                type: "DownloadProgress",
                loaded: event.loaded
            };
            if (event.lengthComputable) {
                progressEvent.total = event.total;
            }
            if (xhr.responseType === "text" && !!xhr.responseText) {
                progressEvent.partialText = xhr.responseText;
            }
            observer.next(progressEvent);
        };
        var onUpProgress = function (event) {
            var progress = {
                type: "UploadProgress",
                loaded: event.loaded
            };
            if (event.lengthComputable) {
                progress.total = event.total;
            }
            observer.next(progress);
        };
        xhr.addEventListener('load', onLoad);
        xhr.addEventListener('error', onError);
        if (config.reportProgress) {
            xhr.addEventListener('progress', onDownProgress);
            if (!!requestData && xhr.upload) {
                xhr.upload.addEventListener('progress', onUpProgress);
            }
        }
        xhr.send(requestData || null);
        observer.next({ type: "Sent" });
        return tearDown;
    });
};
export { xhrAdapter as adapter };
//# sourceMappingURL=xhr.js.map